import duckdb
import fsspec
import json
from .abstract import AbstractCatalog
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.config import RECAP_HOME, settings
from typing import List, Any, Generator
from urllib.parse import urlparse


DEFAULT_URL = f"file://{settings('root_path', RECAP_HOME)}/catalog/recap.duckdb"


class DuckDbCatalog(AbstractCatalog):
    def __init__(
        self,
        connection: duckdb.DuckDBPyConnection,
    ):
        self.connection = connection
        # TODO should try this and catch if read_only=True
        self.connection.execute(
            "CREATE TABLE IF NOT EXISTS catalog "
            "(parent VARCHAR, name VARCHAR, metadata JSON)"
        )
        self.connection.execute(
            "CREATE UNIQUE INDEX IF NOT EXISTS parent_name_idx ON catalog (parent, name)"
        )

    def touch(
        self,
        path: PurePosixPath,
    ):
        path = PurePosixPath('/', path)
        # Reverse to start with root.
        path_stack = list(path.parts)[::-1]
        cwd = '/'

        # Touch all parents to make sure they exist as well
        while len(path_stack):
            cwd = PurePosixPath(cwd, path_stack.pop())

            # PurePosixPath('/').parts returns ('/',). We don't want to touch
            # the root because it doesn't fit the parent/name model that we
            # have.
            if len(cwd.parts) > 1:
                self.connection.begin()

                # TODO An UPSERT would be great here. Doesn't seem DuckDB supports it.
                results = self.connection.execute(
                    "SELECT 1 FROM catalog WHERE parent = ? AND name = ?",
                    [str(cwd.parent), cwd.name]
                )

                if not results.fetchone():
                    doc = self._get_metadata(cwd) or {}
                    self.connection.execute(
                        "INSERT INTO catalog VALUES (?, ?, ?)",
                        [str(cwd.parent), cwd.name, json.dumps(doc)]
                    )

                self.connection.commit()


    def write(
        self,
        path: PurePosixPath,
        type: str,
        metadata: Any,
    ):
        path = PurePosixPath('/', path)

        # TODO An UPSERT would be great here. Doesn't seem DuckDB supports it.
        self.touch(path)
        self.connection.begin()
        doc = self._get_metadata(path)
        updated_doc = (doc or {}) | {type: metadata}
        self.connection.execute(
            (
                "UPDATE catalog SET metadata = ? "
                "WHERE parent = ? AND name = ?"
            ),
            [json.dumps(updated_doc), str(path.parent), path.name]
        )
        self.connection.commit()

    def rm(
        self,
        path: PurePosixPath,
        type: str | None = None,
    ):
        path = PurePosixPath('/', path)
        if not type:
            self.connection.execute(
                (
                    "DELETE FROM catalog WHERE "
                    "parent LIKE ? OR (parent = ? AND name = ?)"
                ),
                [f"{path}%", str(path.parent), path.name]
            )
        else:
            self.connection.begin()
            doc = self._get_metadata(path)
            if doc:
                doc.pop(type, None)
                self.connection.execute(
                    (
                        "UPDATE catalog SET metadata = ? "
                        "WHERE parent = ? AND name = ?"
                    ),
                    [json.dumps(doc), str(path.parent), path.name]
                )
            self.connection.commit()

    def ls(
        self,
        path: PurePosixPath,
    ) -> List[str] | None:
        path = PurePosixPath('/', path)
        self.connection.execute(
            "SELECT name FROM catalog WHERE parent = ?", [str(path)]
        )
        rows = self.connection.fetchall()
        # Get the name from each row
        names = list(map(lambda r: r[0], rows)) if rows else None
        return names # pyright: ignore [reportGeneralTypeIssues]

    def read(
        self,
        path: PurePosixPath,
    ) -> dict[str, Any] | None:
        path = PurePosixPath('/', path)
        return self._get_metadata(path)

    def search(self, query: str) -> List[dict[str, Any]]:
        results = []

        self.connection.execute(
            # ZOMG SQL injection. Should figure out a better way to do this.
            f"SELECT metadata FROM catalog WHERE {query}"
        )

        for row in self.connection.fetchall():
            results.append(json.loads(row[0]))

        return results

    def _get_metadata(self, path: PurePosixPath) -> Any | None:
        self.connection.execute(
            (
                "SELECT metadata FROM catalog "
                "WHERE parent = ? AND name = ?"
            ),
            [str(path.parent), path.name]
        )
        maybe_row = self.connection.fetchone()
        return json.loads(maybe_row[0]) if maybe_row else None # pyright: ignore [reportGeneralTypeIssues]

    @staticmethod
    @contextmanager
    def open(**config) -> Generator['DuckDbCatalog', None, None]:
        url = urlparse(config.get('url', DEFAULT_URL))
        duckdb_options = config.get('duckdb', {})
        duckdb_dir = PurePosixPath(url.path).parent
        try:
            fsspec \
                .filesystem('file', auto_mkdir=True) \
                .mkdirs(duckdb_dir, exist_ok=True)
        except FileExistsError:
            pass
        with duckdb.connect(url.path, **duckdb_options) as c:
            yield DuckDbCatalog(c)



import typer
from fastapi import Body, Depends, FastAPI
from pathlib import PurePosixPath
from typing import Any, List, Generator
from recap.plugins import catalogs
from recap.plugins.catalogs.abstract import AbstractCatalog
from recap.config import settings
from recap.logging import setup_logging


DEFAULT_URL = 'http://localhost:8000'

app = typer.Typer()
fastapp = FastAPI()


def get_catalog() -> Generator[AbstractCatalog, None, None]:
    with catalogs.open(**settings('catalog', {})) as c:
        yield c


# WARN This must go before get_path since get_path is a catch-all.
@fastapp.get("/search")
def query_search(
    query: str,
    catalog: AbstractCatalog = Depends(get_catalog),
) -> List[dict[str, Any]]:
    return catalog.search(query)


@fastapp.get("/{path:path}")
def get_path(
    # TODO Make this a PurePosixPath type. FastAPI is hassling me right now.
    path: str,
    read: bool | None = None,
    catalog: AbstractCatalog = Depends(get_catalog),
) -> List[str] | dict[str, Any]:
    # TODO should probably return a 404 if we get None from storage
    if read:
        return catalog.read(PurePosixPath(path)) or {}
    else:
        return catalog.ls(PurePosixPath(path)) or []


@fastapp.put("/{path:path}")
def put_path(
    # TODO Make this a PurePosixPath type. FastAPI is hassling me right now.
    path: str,
    type: str | None = None,
    metadata: Any = Body(default=None),
    catalog: AbstractCatalog = Depends(get_catalog),
):
    if type and metadata:
        catalog.write(PurePosixPath(path), type, metadata)
    else:
        return catalog.touch(PurePosixPath(path))


@fastapp.delete("/{path:path}")
def delete_path(
    # TODO Make this a PurePosixPath type. FastAPI is hassling me right now.
    path: str,
    type: str | None = None,
    catalog: AbstractCatalog = Depends(get_catalog),
):
    catalog.rm(PurePosixPath(path), type)


@app.command()
def serve():
    """
    Starts a FastAPI server that exposes a catalog API over HTTP/JSON.
    """

    import uvicorn

    uvicorn.run(
        fastapp,
        log_config=setup_logging(),
        **settings('api', {}),
    )



import logging
import sqlalchemy as sa
from .abstract import AbstractAnalyzer
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.plugins.browsers.db import DatabasePath, DatabaseBrowser
from typing import Any, Generator, List


log = logging.getLogger(__name__)


class AbstractTableAnalyzer(AbstractAnalyzer):
    def __init__(
        self,
        engine: sa.engine.Engine,
    ):
        self.engine = engine

    @staticmethod
    def analyzable(url: str) -> bool:
        # TODO there's probably a better way to do this.
        # Seems like SQLAlchemy should have a method to check dialects.
        try:
            sa.create_engine(url)
            return True
        except Exception as e:
            log.debug('Unanalyzable. Create engine failed for url=%s', url)
            return False

    def analyze(self, path: PurePosixPath) -> dict[str, Any]:
        database_path = DatabasePath(path)
        schema = database_path.schema
        table = database_path.table
        if schema and table:
            is_view = path.parts[3] == 'views'
            return self.analyze_table(schema, table, is_view)
        return {}

    @abstractmethod
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        raise NotImplementedError

    @classmethod
    @contextmanager
    def open(cls, **config) -> Generator['AbstractTableAnalyzer', None, None]:
        assert 'url' in config, \
            f"Config for {cls.__name__} is missing `url` config."
        engine = sa.create_engine(config['url'])
        yield cls(engine)


class TableLocationAnalyzer(AbstractTableAnalyzer):
    def __init__(
        self,
        root: PurePosixPath,
        engine: sa.engine.Engine,
    ):
        self.root = root
        self.engine = engine
        self.database = root.parts[2]
        self.instance = root.parts[4]

    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        if schema and table:
            table_or_view = 'view' if is_view else 'table'
            return {
                'location': {
                    'database': self.database,
                    'instance': self.instance,
                    'schema': schema,
                    table_or_view: table,
                }
            }
        return {}

    @classmethod
    @contextmanager
    def open(cls, **config) -> Generator['TableLocationAnalyzer', None, None]:
        assert 'url' in config, \
            f"Config for {cls.__name__} is missing `url` config."
        engine = sa.create_engine(config['url'])
        root = DatabaseBrowser.root(**config)
        yield TableLocationAnalyzer(root, engine)


class TableColumnAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        results = {}
        columns = sa.inspect(self.engine).get_columns(table, schema)
        for column in columns:
            if column.get('comment', None) is None:
                del column['comment']
            try:
                generic_type = column['type'].as_generic()
                # Strip length/precision to make generic strings more generic.
                if isinstance(generic_type, sa.sql.sqltypes.String):
                    generic_type.length = None
                elif isinstance(generic_type, sa.sql.sqltypes.Numeric):
                    generic_type.precision = None
                    generic_type.scale = None
                column['generic_type'] = str(generic_type)
            except NotImplementedError as e:
                # Unable to convert. Probably a weird type like PG's OID.
                log.debug(
                    'Unable to get generic type for table=%s.%s column=%s',
                    schema,
                    table,
                    column.get('name', column),
                    exc_info=e,
                )
            # The `type` field is not JSON encodable; convert to string.
            column['type'] = str(column['type'])
            column_name = column['name']
            del column['name']
            results[column_name] = column
        return {'columns': results} if results else {}


class TableIndexAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        indexes = {}
        index_dicts = sa.inspect(self.engine).get_indexes(table, schema)
        for index_dict in index_dicts:
            indexes[index_dict['name']] = {
                'columns': index_dict.get('column_names', []),
                'unique': index_dict['unique'],
            }
        return {'indexes': indexes} if indexes else {}


class TablePrimaryKeyAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        pk_dict = sa.inspect(self.engine).get_pk_constraint(table, schema)
        return {'primary_key': pk_dict} if pk_dict else {}


class TableForeignKeyAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        fk_dict = sa.inspect(self.engine).get_foreign_keys(table, schema)
        return {'foreign_keys': fk_dict} if fk_dict else {}


class TableViewDefinitionAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        if not is_view:
            return {}
        # TODO sqlalchemy-bigquery doesn't work right with this API
        # https://github.com/googleapis/python-bigquery-sqlalchemy/issues/539
        if self.engine.dialect.name == 'bigquery':
            table = f"{schema}.{table}"
        def_dict = sa.inspect(self.engine).get_view_definition(table, schema)
        return {'view_definition': def_dict} if def_dict else {}


class TableCommentAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        try:
            comment = sa.inspect(self.engine).get_table_comment(table, schema)
            comment_text = comment.get('text')
            return {'comment': comment_text} if comment_text else {}
        except NotImplementedError as e:
            log.debug(
                'Unable to get comment for table=%s.%s',
                schema,
                table,
                exc_info=e,
            )
            return {}


class TableAccessAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        with self.engine.connect() as conn:
            results = {}
            try:
                rows = conn.execute(
                    "SELECT * FROM information_schema.role_table_grants "
                    "WHERE table_schema = %s AND table_name = %s",
                    schema,
                    table,
                )
                for row in rows.all():
                    privilege_type = row['privilege_type']
                    user_grants: dict[str, Any] = results.get(row['grantee'], {
                        'privileges': [],
                        'read': False,
                        'write': False,
                    })
                    user_grants['privileges'].append(privilege_type)
                    if privilege_type == 'SELECT':
                        user_grants['read'] = True
                    if privilege_type in ['INSERT', 'UPDATE', 'DELETE', 'TRUNCATE']:
                        user_grants['write'] = True
                    results[row['grantee']] = user_grants
            except Exception as e:
                # TODO probably need a more tightly bound exception here
                # We probably don't have access to the information_schema, so
                # skip it.
                log.debug(
                    'Unable to fetch access for table=%s.%s',
                    schema,
                    table,
                    exc_info=e,
                )
            return {'access': results} if results else {}


class TableProfileAnalyzer(AbstractTableAnalyzer):
    def analyze_table(
        self,
        schema: str,
        table: str,
        is_view: bool = False
    ) -> dict[str, Any]:
        column_analyzer = TableColumnAnalyzer(self.engine)
        # TODO This is very proof-of-concept...
        # TODO ZOMG SQL injection attacks all over!
        # TODO Is db.Table().select the right way to paramaterize tables?
        columns = column_analyzer \
            .analyze_table(schema, table) \
            .get('columns', {})
        sql_col_queries = ''
        numeric_types = [
            'BIGINT', 'FLOAT', 'INT', 'INTEGER', 'NUMERIC', 'REAL', 'SMALLINT'
        ]
        date_types = [
            'DATE', 'DATETIME', 'TIMESTAMP'
        ]
        # TODO Excluding 'JSON' because PG's 'JSONB' doesn't have LENGTH()
        string_types = [
            'CHAR', 'CLOB', 'NCHAR', 'NVARCHAR', 'TEXT', 'VARCHAR'
        ]
        binary_types = [
            'BLOB', 'VARBINARY'
        ]
        stat_types = [
            'min',
            'max',
            'average',
            'sum',
            'distinct',
            'nulls',
            'zeros',
            'negatives',
            'min_length',
            'max_length',
            'empty_strings',
            'unix_epochs',
        ]

        with self.engine.connect() as conn:
            varchar_type = 'VARCHAR'

            # BigQuery doesn't havt FLOAT or VARCHAR, so use its type.
            # TODO SQLAlchemy should expose a dialect type for a generic type.
            if conn.dialect.name == 'bigquery':
                float_type = 'FLOAT64'
                varchar_type = 'STRING'
            elif conn.dialect.name == 'snowflake':
                conn.execute("ALTER SESSION SET QUOTED_IDENTIFIERS_IGNORE_CASE = TRUE")

            for column_name, column in columns.items():
                generic_type = column.get('generic_type')
                quoted_column_name = f'"{column_name}"'
                if conn.dialect.name in ['bigquery', 'mysql']:
                    quoted_column_name = f'`{column_name}`'
                if generic_type in numeric_types:
                    # TODO add approx median and quantiles
                    # TODO can we use a STRUCT or something here?
                    sql_col_queries += f"""
                        , MIN({quoted_column_name}) AS min_{column_name}
                        , MAX({quoted_column_name}) AS max_{column_name}
                        , AVG({quoted_column_name}) AS average_{column_name}
                        , SUM({quoted_column_name}) AS sum_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS {column_name}_distinct
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} = 0 THEN 1 ELSE 0 END) AS zeros_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} < 0 THEN 1 ELSE 0 END) AS negatives_{column_name}
                    """
                if generic_type in string_types:
                    sql_col_queries += f"""
                        , MIN(LENGTH({quoted_column_name})) AS min_length_{column_name}
                        , MAX(LENGTH({quoted_column_name})) AS max_length_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} LIKE '' THEN 1 ELSE 0 END) AS empty_strings_{column_name}
                    """
                if generic_type in binary_types:
                    sql_col_queries += f"""
                        , MIN(LENGTH({quoted_column_name})) AS min_length_{column_name}
                        , MAX(LENGTH({quoted_column_name})) AS max_length_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                    """
                if generic_type in date_types:
                    sql_col_queries += f"""
                        , CAST(MIN({quoted_column_name}) AS {varchar_type}) AS min_{column_name}
                        , CAST(MAX({quoted_column_name}) AS {varchar_type}) AS max_{column_name}
                        , COUNT(DISTINCT {quoted_column_name}) AS distinct_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} IS NULL THEN 1 ELSE 0 END) AS nulls_{column_name}
                        , SUM(CASE WHEN {quoted_column_name} = TIMESTAMP '1970-01-01 00:00:00' THEN 1 ELSE 0 END) AS unix_epochs_{column_name}
                    """

            quoted_schema = f'"{schema}"'
            quoted_table = f'"{table}"'
            if conn.dialect.name in ['bigquery', 'mysql']:
                quoted_schema = f'`{schema}`'
                quoted_table = f'`{table}`'

            sql = f"""
                SELECT
                    COUNT(*) AS count {sql_col_queries}
                FROM
                    {quoted_schema}.{quoted_table} t
            """

            rows = conn.execute(sql)
            row = dict(rows.first() or {})
            results = {}

            for column_name in columns.keys():
                col_stats = results.get(column_name, {'count': row['count']})
                for stat_type in stat_types:
                    stat_name = f"{stat_type}_{column_name}"
                    if stat_name in row:
                        stat_value = row[stat_name]
                        # JSON encoder can't handle decimal.Decimal
                        import decimal
                        if isinstance(row[stat_name], decimal.Decimal):
                            stat_value = float(row[stat_name])
                        col_stats[stat_type] = stat_value
                results[column_name] = col_stats

            return {'profile': results}



import httpx
from .abstract import AbstractCatalog
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.plugins.commands.serve import DEFAULT_URL
from typing import Any, List, Generator


class RecapCatalog(AbstractCatalog):
    """
    RecapCatalog makes HTTP requests to a Recap server and returns the results.
    """

    def __init__(
        self,
        client: httpx.Client,
    ):
        self.client = client

    def touch(
        self,
        path: PurePosixPath,
    ):
        self.client.put(str(path))

    def write(
        self,
        path: PurePosixPath,
        type: str,
        metadata: Any,
    ):
        params = {'type': type} if type else None
        self.client.put(str(path), params=params, json=metadata)

    def rm(
        self,
        path: PurePosixPath,
        type: str | None = None,
    ):
        params = {'type': type} if type else None
        self.client.delete(str(path), params=params)

    def ls(
        self,
        path: PurePosixPath,
    ) -> List[str] | None:
        return self.client.get(str(path)).json()

    def read(
        self,
        path: PurePosixPath,
    ) -> dict[str, Any] | None:
        return self.client.get(str(path), params={'read': True}).json()

    def search(
        self,
        query: str,
    ) -> List[dict[str, Any]]:
        return self.client.get('/search', params={'query': query}).json()

    @staticmethod
    @contextmanager
    def open(**config) -> Generator['RecapCatalog', None, None]:
        url = config.get('url', DEFAULT_URL)
        with httpx.Client(base_url=url) as client:
            yield RecapCatalog(client)



from .abstract import AbstractCatalog
from contextlib import contextmanager
from typing import Generator


@contextmanager
def open(**config) -> Generator['AbstractCatalog', None, None]:
    from recap.plugins import load_catalog_plugins
    type = config.get('type', 'duckdb')
    catalog_plugins = load_catalog_plugins()
    catalog_plugin_cls = catalog_plugins.get(type)
    assert catalog_plugin_cls, \
        f"Unable to find catalog plugin={type}"
    with catalog_plugin_cls.open(**config) as c:
        yield c



import fsspec
import json
import logging
import pyjq
from .abstract import AbstractCatalog
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.config import RECAP_HOME, settings
from typing import Any, List, Generator
from urllib.parse import urlparse


DEFAULT_URL = f"file://{settings('root_path', RECAP_HOME)}/catalog"
log = logging.getLogger(__name__)


class FilesystemCatalog(AbstractCatalog):
    def __init__(
        self,
        root: PurePosixPath,
        fs: fsspec.AbstractFileSystem,
    ):
        self.root = root
        self.fs = fs

    def touch(
        self,
        path: PurePosixPath,
    ):
        full_path = PurePosixPath(self.root, str(path).strip('/'))
        self.fs.mkdirs(full_path, exist_ok=True)

    def write(
        self,
        path: PurePosixPath,
        type: str,
        metadata: Any,
    ):
        full_path = PurePosixPath(self.root, str(path).strip('/'), f'{type}.json')
        if not self.fs.exists(full_path.parent):
            self.fs.mkdirs(full_path.parent, exist_ok=True)
        with self.fs.open(full_path, 'w+') as f:
            json.dump(metadata, f) # pyright: ignore [reportGeneralTypeIssues]

    def rm(
        self,
        path: PurePosixPath,
        type: str | None = None,
    ):
        full_path = PurePosixPath(self.root, str(path).strip('/'))
        if type:
            full_path = PurePosixPath(full_path, f"{type}.json")
        try:
            self.fs.rm(full_path, recursive=True)
        except FileNotFoundError:
            # File is already deleted
            # TODO Maybe we should raise a StorageException here?
            pass

    def ls(
        self,
        path: PurePosixPath,
    ) -> List[str] | None:
        try:
            # I thought there might be a vulnerability where users could
            # specify '/..' to read arbitrary parts of the filesystem, but
            # PurePosixPath seems to handle that properly, and stops at
            # self.root. So, just strip leading and trailing slashes to prevent
            # overriding self.root.
            dir = PurePosixPath(self.root, str(path).strip('/'))
            children = self.fs.ls(dir, detail=True) if self.fs.exists(dir) else []
            # Remove files since we're listing.
            children = filter(lambda c: c['type'] == 'directory', children)
            # Remove full path and only show immediate child names.
            children = map(lambda c: PurePosixPath(c['name']).name, children)
            return list(children)
        except FileNotFoundError:
            return None

    def read(
        self,
        path: PurePosixPath,
    ) -> dict[str, Any] | None:
        dir = PurePosixPath(self.root, str(path).strip('/'))
        doc = {}
        try:
            for child in self.fs.ls(dir, detail=True):
                if child['type'] == 'file':
                    # Strip .json from the filename and remove prefix path
                    type = PurePosixPath(child['name']).with_suffix('').name
                    with self.fs.open(child['name'], 'r') as f:
                        doc[type] = json.load(f)
            return doc
        except FileNotFoundError:
            # File is already deleted
            # TODO Maybe we should raise a StorageException here?
            return None

    def search(self, query: str) -> List[dict[str, Any]]:
        results = []
        path_stack = [PurePosixPath('/')]

        while path_stack:
            path = path_stack.pop()
            doc = self.read(PurePosixPath(path)) or {}

            # If the doc matches the query, add it to the results
            if pyjq.first(query, doc):
                results.append(doc)

            # Now add any children to the stack for processing
            children = self.ls(path) or []
            children = map(lambda c: PurePosixPath(path, c), children)
            path_stack.extend(children)

        return results

    @staticmethod
    @contextmanager
    def open(**config) -> Generator['FilesystemCatalog', None, None]:
        url = urlparse(config.get('url', DEFAULT_URL))
        storage_options = config.get('fs', {})
        fs = fsspec.filesystem(
            url.scheme,
            **storage_options,
            # TODO This should move to the filesystem storage config
            auto_mkdir=True)
        yield FilesystemCatalog(
            PurePosixPath(url.path),
            fs,
        )



import logging
import sqlalchemy as sa
from contextlib import contextmanager
from .abstract import AbstractBrowser
from pathlib import PurePosixPath
from typing import Callable, Generator, List
from urllib.parse import urlparse


log = logging.getLogger(__name__)


class DatabasePath:
    """
    Helper that exposes schema and table (or view) from a database path.
    """

    def __init__(self, path: PurePosixPath):
        self.path = path
        self.schema = path.parts[2] if len(path.parts) > 2 else None
        self.table = path.parts[4] if len(path.parts) > 4 else None


class DatabaseBrowser(AbstractBrowser):
    """
    A browser that lists database objects. DatabaseBrowser uses SQLAlchemy and
    its supported dialects (https://docs.sqlalchemy.org/en/13/dialects/).

    DatabaseBrowser follows this directory sturcture:

        /databases/<scheme>/instances/<instance>/schemas/<schema>/tables/<table>
        /databases/<scheme>/instances/<instance>/schemas/<schema>/views/<view>

    The scheme is derived from the DB URL (e.g.
    postgres://user:pass@localhost/my_db would have scheme="postgres"). The
    instance defaults to the hostname, but can be overridden via the `name`
    config. Schema is whatever SQLAlchemy returns from `get_schema_names`.
    Ditto for table and view.

    This model does not make the distinction between schemas and databases as
    defined in standard information_schema formats
    (https://en.wikipedia.org/wiki/Information_schema). PostgreSQL, in
    particular, is a little weird because it has both (the schema is usually
    `public`).
    """

    def __init__(
        self,
        engine: sa.engine.Engine,
    ):
        self.url = str(engine.url)
        self.engine = engine

    def children(self, path: PurePosixPath) -> List[str]:
        num_parts = len(path.parts)
        database_path = DatabasePath(path)
        # TODO Should use a lookup table here for better performance?
        if num_parts == 1:
            # /
            return ['schemas']
        elif num_parts == 2 and path.parts[1] == 'schemas':
            # /schemas
            return self.schemas()
        elif num_parts == 3:
            # /schemas/<some_schema>
            return ['tables', 'views']
        elif num_parts == 4 and database_path.schema:
            if path.parts[3] == 'tables':
                # /schemas/<some_schema>/tables
                return self.tables(database_path.schema)
            if path.parts[3] == 'views':
                # /schemas/<some_schema>/views
                return self.views(database_path.schema)
        return []

    def schemas(self) -> List[str]:
        """
        :returns: All schema names in a database. In PostgreSQL, this is
            usually just `public`. In MySQL and others, it's usually a list of
            all database names.
        """

        return sa.inspect(self.engine).get_schema_names()

    def tables(self, schema: str) -> List[str]:
        """
        :returns: All table names in a schema.
        """

        return self._tables_or_views(
            schema,
            sa.inspect(self.engine).get_table_names,
        )

    def views(self, schema: str) -> List[str]:
        """
        :returns: All view names in a schema.
        """
        return self._tables_or_views(
            schema,
            sa.inspect(self.engine).get_view_names,
        )

    def _tables_or_views(
        self,
        schema: str,
        get_method: Callable[[str], List[str]],
    ) -> List[str]:
        """
        Helper function that gets returns all tables or views for a given
        schema. This method exists because some DBs return `<schema>.<table>`
        and others just return `<table>`. To keep things standard, we strip out
        the `<schema>.` prefix if it exists.

        :param get_method: A SQLAlchemy inspection method; either
            (`get_table_names` or `get_schema_names`).
        :returns: All views or tables in a schema.
        """

        results = []
        try:
            for table_or_view in get_method(schema):
                # Stripe schema name from the table/view name. Some dialects
                # include the schema name as part of the table/view. Let's keep
                # things consistent.
                if table_or_view.startswith(f"{schema}."):
                    table_or_view = table_or_view[len(schema) + 1:]
                results.append(table_or_view)
        except Exception as e:
            # Just optimistically try, and ignore if we can't get info.
            # Easier than trying to figure out if permission exists.
            log.debug(
                'Unable to fetch tables or views for schema=%s',
                schema,
                exc_info=e,
            )
        return results

    @staticmethod
    def root(**config) -> PurePosixPath:
        """
        :returns: A path of the format /databases/<scheme>/instances/<instance>
        """

        assert 'url' in config, \
            f"No url defined for browser config={config}"
        parsed_url = urlparse(config['url'])
        # Given `posgrestql+psycopg2://foo:bar@baz/some_db`, return `postgresql`.
        infra = parsed_url.scheme.split('+')[0]
        # Given `posgrestql+psycopg2://foo:bar@baz/some_db`, return `baz`.
        default_instance = parsed_url.netloc.split('@')[-1]
        instance = config['name'] if 'name' in config else default_instance
        return PurePosixPath(
            '/',
            'databases', infra,
            'instances', instance,
        )

    @staticmethod
    def browsable(url: str) -> bool:
        """
        :returns: True if a SQLAlchemy engine can be created, else False.
        """

        # TODO there's probably a better way to do this.
        # Seems like SQLAlchemy should have a method to check dialects.
        try:
            sa.create_engine(url)
            return True
        except Exception as e:
            log.debug('Unbrowsable. Create engine failed for url=%s', url)
            return False

    @staticmethod
    @contextmanager
    def open(**config) -> Generator['DatabaseBrowser', None, None]:
        assert 'url' in config, \
            f"No url defined for browser config={config}"
        yield DatabaseBrowser(sa.create_engine(config['url']))



from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import PurePosixPath
from typing import Generator, List


class AbstractBrowser(ABC):
    """
    The abstract class for all browsers. Recap uses a browser abstraction to
    deal with different data infrastructure in a standard way.

    Browsers map infrastructure objects into a standard directory format. A
    different browser is used for each type of infrastructure.

    A KafkaBrowser might list directories like this:

        /
        /streaming
        /streaming/kafka
        /streaming/kafka/instances
        /streaming/kafka/instances/log-cluster
        /streaming/kafka/instances/log-cluster/streams
        /streaming/kafka/instances/log-cluster/streams/access-logs
        /streaming/kafka/instances/log-cluster/streams/error-logs
    """

    url: str = NotImplemented
    """
    The base URL for the infrastructure that this browser is browsing.
    Something like 'postgresql://user:pass@localhost/my_db`.
    """

    @staticmethod
    @abstractmethod
    def browsable(url: str) -> bool:
        """
        Determine if this browser can browse a given infrastructure instance.
        In plain english, `browsable` answers the question "Can FooBrowser
        return directory lists for 'postgresql://user:pass@localhost/some_db'"?

        `browsable` is called before the browser is instantiated. If it
        returns False, the browser is skipped.

        :param url: The instance URL that will be browsed. This URL points to
            the base of the instance; it doesn't include a path to specific
            data. For example, it would look like
            "postgresql://user:pass@localhost/some_db", rather than
            "postgresql://user:pass@localhost/some_db/some_schema/some_table".

            `browsable` will be called once, then `children()` will be called
            multiple times, once for each path to data.
        :returns: True if the browser is compatible with the given URL, else
            False.
        """

        raise NotImplementedError

    @abstractmethod
    def children(self, path: PurePosixPath) -> List[str]:
        """
        Given a path, returns its children. Using the example above,
        path="/streaming/kafka/instances/log-cluster/streams" would return:

            ["access-logs", "error-logs"]
        """

        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def root(**config) -> PurePosixPath:
        """
        Returns the root directory for this browser. Using the example above,
        root() would return:

            /streaming/kafka/instances/log-cluster
        """

        raise NotImplementedError

    @staticmethod
    @contextmanager
    @abstractmethod
    def open(**config) -> Generator['AbstractBrowser', None, None]:
        """
        Creates and returns a browser using the supplied config.
        """

        raise NotImplementedError



import typer
from recap.plugins import catalogs
from recap.config import settings
from rich import print_json
from pathlib import PurePosixPath


app = typer.Typer(help="Read and search the data catalog.")


@app.command()
def search(query: str):
    """
    Searches the data catalog. The query string syntax is dependent on the data
    catalog.
    """

    with catalogs.open(**settings('catalog', {})) as c:
        results = c.search(query)
        print_json(data=results, sort_keys=True)


@app.command("list")
def list_(
    path: str = typer.Argument('/')
):
    """
    Lists a data catalog directory's children.
    """

    with catalogs.open(**settings('catalog', {})) as c:
        results = sorted(c.ls(PurePosixPath(path)) or [])
        print_json(data=results)


@app.command()
def read(
    path: str
):
    """
    Prints metadata from a path in the data catalog.
    """

    with catalogs.open(**settings('catalog', {})) as c:
        results = c.read(PurePosixPath(path))
        print_json(data=results, sort_keys=True)



from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import PurePosixPath
from typing import Any, Generator


class AbstractAnalyzer(ABC):
    """
    The abstract class for all analyzers. Analyzers are responsible for
    inspecting data and returning metadata.
    """

    @staticmethod
    @abstractmethod
    def analyzable(url: str) -> bool:
        """
        Determine if this analyzer can analyze data from a given infrastructure
        instance. In plain english, `analyzable` answers the question "Can
        FooAnalyzer collect metadata from
        'postgresql://user:pass@localhost/some_db'"?

        `analyzable` is called before the analyzer is instantiated. If it
        returns False, the analyzer is skipped.

        :param url: The instance URL that to be analyzed. This URL points to
            the base of the instance; it doesn't include a path to specific
            data. For example, it would look like
            "postgresql://user:pass@localhost/some_db", rather than
            "postgresql://user:pass@localhost/some_db/some_schema/some_table".

            `analyzable` will be called once, then `analyze()` will be called
            multiple times, once for each path to data.
        :returns: True if the analyzer is compatible with the given URL, else
            False.
        """

        raise NotImplementedError

    @abstractmethod
    def analyze(self, path: PurePosixPath) -> dict[str, Any]:
        """
        Analyze a path for an infrastructure instance. Only the path is
        specified because the URL for the instance is passed in via the config
        in `open()`.

        :returns: Metadata dictionary of the format {"metadata_type": Any}.
            This data gets serialized as JSON in the catalog.
        """

        raise NotImplementedError

    @staticmethod
    @contextmanager
    @abstractmethod
    def open(**config) -> Generator['AbstractAnalyzer', None, None]:
        """
        Creates and returns an analyzer using the supplied config.
        """

        raise NotImplementedError
